"""
This file implements the basic Dataset class.
"""
import json # for pretty printing dict
import warnings
from textwrap import dedent
from math import factorial

warnings.filterwarnings("ignore", message="divide by zero encountered in true_divide")

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from pysprint.core.bases.dataset_base import DatasetBase, C_LIGHT
from pysprint.core.preprocess import savgol, find_peak, convolution, cut_data, cwt
from pysprint.mpl_tools.normalize import DraggableEnvelope
from pysprint.utils.exceptions import *
from pysprint.mpl_tools.spp_editor import SPPEditor
from pysprint.utils import MetaData, run_from_ipython, find_nearest

__all__ = ['Dataset']


class Dataset(DatasetBase):
	"""
	Base class for the evaluating methods.
	"""
	meta = MetaData("""Additional info about the dataset""", copy=False)

	def __init__(self, x, y, ref=None, sam=None, meta=None):
		
		super().__init__()

		self.x = x
		self.y = y
		if ref is None:
			self.ref = []
		else:
			self.ref = ref 
		if sam is None:
			self.sam = []
		else:
			self.sam = sam
		self._is_normalized = False
		if not isinstance(self.x, np.ndarray):
			try:
				self.x = np.array(self.x).astype(float)
			except ValueError:
				raise DatasetError('Invalid type of data')
		if not isinstance(self.y, np.ndarray):
			try:
				self.y = np.array(self.y).astype(float)
			except ValueError:
				raise DatasetError('Invalid type of data')
		if not isinstance(self.ref, np.ndarray):
			try:
				self.ref = np.array(self.ref).astype(float)
			except ValueError:
				pass # just ignore invalid arms
		if not isinstance(self.sam, np.ndarray):
			try:
				self.sam = np.array(self.sam).astype(float)
			except ValueError:
				pass # just ignore invalid arms

		if len(self.ref) == 0:
			self.y_norm = self.y
			self._is_normalized = self._ensure_norm()

		else:
			self.y_norm = (self.y - self.ref - self.sam) / (2 * np.sqrt(self.sam * self.ref))
			self._is_normalized = True
		
		self.plotwidget = plt
		self.xmin = None
		self.xmax = None
		self.probably_wavelength = None
		self._check_domain()
		
		if meta is not None:
			self.meta = meta

		self._delay = None
		self._positions = None

		self._dispersion_array = None


	def phase_plot(self, exclude_GD=False):
		if not np.all(self._dispersion_array):
			raise ValueError('Dispersion must be calculated before plotting the phase.')

		coefs = np.array(
			[self._dispersion_array[i]/factorial(i+1) for i in range(len(self._dispersion_array))]
			)

		if exclude_GD:
			coefs[0] = 0

		phase_poly = np.poly1d(coefs[::-1], r=False)
		
		self.plotwidget.plot(self.x, phase_poly(self.x))
		self.plotwidget.grid()
		self.plotwidget.ylabel('$\Phi\, [rad]$')
		self.plotwidget.xlabel('$\omega \,[PHz]$')
		self.plotwidget.show()
		
	@property
	def delay(self):
		return self._delay

	@delay.setter
	def delay(self, value):
		self._delay = value

	@property
	def positions(self):
		return self._positions

	@positions.setter
	def positions(self, value):
		self._positions = value


	def _ensure_norm(self):
		idx = np.where((self.y_norm > 2))
		val = len(idx[0]) / len(self.y_norm)
		if val > 0.015: # this is a custom threshold, which often works..
			return False
		return True

	def scale_up(self):
		'''
		If the interferogram is normalized to [0, 1] interval, scale up to [-1, 1]
		with easy algerbra.. Just in case you need comparison, or any other purpose.
		'''
		self.y_norm = (self.y_norm - 0.5) * 2
		self.y = (self.y - 0.5) * 2

	def GD_lookup(self, reference_point=2.355, engine='cwt', silent=False, **kwargs):
		'''
		Quick GD lookup: it finds extremal points near the `reference_point` and returns
		an avarage value of 2*np.pi divided by distances between consecutive minimal or
		maximal values. Since it's relying on peak detection, the results may be irrelevant
		in some cases. If the parent class is `~pysprint.CosFitMethod`, then it will set the
		predicted value as inital parameter for fitting.
		'''
		if engine not in ('cwt', 'normal'):
			raise ValueError('Engine must be `cwt` or `normal`.')

		if engine == 'cwt':
			width = kwargs.pop('width', 35)
			floor_thres = kwargs.pop('floor_thres', 0.05)
			x_min, _, x_max, _ = self.detect_peak_cwt(width=width, floor_thres=floor_thres)

			#just validation
			_ = kwargs.pop('pmin', 0.1)
			_ = kwargs.pop('pmax', 0.1)
			_ = kwargs.pop('threshold', 0.35)

		else:
			pmin = kwargs.pop('pmin', 0.1)
			pmax = kwargs.pop('pmax', 0.1)
			threshold = kwargs.pop('threshold', 0.35)
			x_min, _, x_max, _ = self.detect_peak(pmin=pmin, pmax=pmax, threshold=threshold)

			# just validation
			_ = kwargs.pop('width', 10)
			_ = kwargs.pop('floor_thres', 0.05)

		if kwargs:
			raise TypeError(f'Invalid argument:{kwargs}')

		try:
			closest_val, idx1 = find_nearest(x_min, reference_point)
			m_closest_val, m_idx1 = find_nearest(x_max, reference_point)
		except ValueError:
			if not silent:
				print('Prediction failed.\nSkipping.. ')
			return
		try:
			truncated = np.delete(x_min, idx1)
			second_closest_val, _ = find_nearest(truncated, reference_point)
		except IndexError:
			if not silent:
				print('Prediction failed.\nSkipping.. ')
			return
		try:
			m_truncated = np.delete(x_max, m_idx1)
			m_second_closest_val, _ = find_nearest(m_truncated, reference_point)
		except IndexError:
			if not silent:
				print('Prediction failed.\nSkipping.. ')
			return

		lowguess = 2*np.pi/np.abs(closest_val-second_closest_val)
		highguess = 2*np.pi/np.abs(m_closest_val-m_second_closest_val)

		#  estimate the GD with that
		if hasattr(self, 'params'):
			self.params[3] = (lowguess + highguess) / 2

		if not silent:
			print(f'The predicted GD is ± {((lowguess + highguess) / 2):.5f} fs based on reference point of {reference_point}.')

	def _safe_cast(self):
		'''
		Return a copy of key attributes in order to prevent inplace modification.
		'''
		x, y, ref, sam = np.copy(self.x), np.copy(self.y), np.copy(self.ref), np.copy(self.sam)
		return x, y, ref, sam

	@staticmethod
	def wave2freq(value):
		'''
		Switches values between wavelength and angular frequency.
		'''
		return (2*np.pi*C_LIGHT)/value

	_dispatch = wave2freq.__func__

	@staticmethod
	def freq2wave(value):
		'''
		Switches values between angular frequency and wavelength.
		'''
		return Dataset._dispatch(value)

	def _check_domain(self):
		"""
		Checks the domain of data just by looking at x axis' minimal value.
		Units are obviously not added yet, we work in nm and PHz...
		"""
		if min(self.x) > 50:
			self.probably_wavelength = True
		else:
			self.probably_wavelength = False

	@classmethod
	def parse_raw(cls, basefile, ref=None, sam=None, skiprows=8,
		decimal=',', sep=';', meta_len=5):
		'''
		Dataset object alternative constructor. Helps to load in data just by giving the filenames
		in the target directory.

		Parameters:
		----------
		basefile: `str`
			base interferogram
			file generated by the spectrometer

		ref: `str`, optional
			reference arm's spectra
			file generated by the spectrometer

		sam: `str`, optional
			sample arm's spectra
			file generated by the spectrometer

		skiprows: `int`, optional
			Skip rows at the top of the file. Default is `8`.

		sep: `str`, optional
			The delimiter in the original interferogram file.
			Default is `;`.

		decimal: `str`, optional
			Character recognized as decimal separator in the original dataset. 
			Often `,` for European data.
			Default is `,`.

		meta_len: `int`, optional
			The first `n` lines in the original file containing the meta information
			about the dataset. It is parsed to be dict-like. If the parsing fails,
			a new entry will be created in the dictionary with key `unparsed`.
			Default is `5`.
		'''
		if skiprows < meta_len:
			warnings.warn(f'Skiprows is currently {skiprows}, but meta information is set to {meta_len} lines. This implies that either one is probably wrong.', PySprintWarning)
		with open(basefile) as file:
			comm = next(file).strip('\n').split('-')[-1].lstrip(' ')
			additional = (next(file).strip('\n').strip('\x00').split(':') for _ in range(1, meta_len))
			if meta_len != 0:
				cls.meta = {'comment': comm}
			try:
				for info in additional:
					cls.meta[info[0]] = info[1]
			except IndexError:
				cls.meta['unparsed'] = str(list(additional))
		df = pd.read_csv(basefile, skiprows=skiprows, sep=sep, decimal=decimal, usecols=[0,1], names=['x', 'y'])
		if (ref is not None and sam is not None):
			r = pd.read_csv(ref, skiprows=skiprows, sep=sep, decimal=decimal, usecols=[0,1], names=['x', 'y'])
			s = pd.read_csv(sam, skiprows=skiprows, sep=sep, decimal=decimal, usecols=[0,1], names=['x', 'y'])
			return cls(df['x'].values, df['y'].values, r['y'].values, s['y'].values)
		return cls(df['x'].values, df['y'].values)

	def __str__(self):
		return self.__repr__()

	def __repr__(self):
		if isinstance(self._delay, np.ndarray):
			pprint_delay = self._delay[0]
		elif isinstance(self._delay, float) or isinstance(self._delay, int):
			pprint_delay = self._delay
		else:
			pprint_delay = '-'
		string = dedent(f'''
		{type(self).__name__} object

		Parameters
		----------
		Datapoints: {len(self.x)}
		Normalized: {self._is_normalized}
		Predicted domain: {'wavelength' if self.probably_wavelength else 'frequency'}
		Delay value: {(str(pprint_delay) + ' fs') if np.all(self._delay) else 'Not given'}
		SPP position(s): {self._positions if np.all(self._positions) else 'Not given'}

		Metadata extracted from file
		----------------------------
		{json.dumps(self.meta, indent=8)}''')
		return string

	@property
	def data(self):
		"""
		Returns the *current* dataset as `pandas.DataFrame`.
		"""
		if self._is_normalized:
			try:
				self._data = pd.DataFrame({
				'x': self.x,
				'y': self.y,
				'sample': self.sam,
				'reference': self.ref,
				'y_normalized': self.y_norm
					})
			except ValueError:
				self._data = pd.DataFrame({
					'x': self.x,
					'y': self.y,
					})
		else:
			self._data = pd.DataFrame({
				'x': self.x,
				'y': self.y,
				})
		return self._data

	@property
	def is_normalized(self):
		"""
		Retuns whether the dataset is normalized.
		"""
		return self._is_normalized
	
	def chdomain(self):
		""" Changes from wavelength [nm] to ang. freq. [PHz] domain and vica versa."""
		self.x = (2*np.pi*C_LIGHT)/self.x
		self._check_domain()
		if type(self).__name__ == 'FFTMethod':
			self.original_x = self.x


	def detect_peak_cwt(self, width, floor_thres=0.05):
		x, y, ref, sam = self._safe_cast()
		xmax, ymax, xmin, ymin = cwt(x, y, ref, sam, width=width, floor_thres=floor_thres)
		return xmax, ymax, xmin, ymin

	def savgol_fil(self, window=5, order=3):
		"""
		Applies Savitzky-Golay filter on the dataset.

		Parameters:
		----------
		window: `int`
			Length of the convolutional window for the filter.
			Default is `10`.

		order: `int`
			Degree of polynomial to fit after the convolution.
			If not odd, it's incremented by 1. Must be lower than window.
			Usually it's a good idea to stay with a low degree, e.g 3 or 5.
			Default is 3.

		Notes:
		------
		If arms were given, it will merge them into the `self.y` and `self.y_norm` variables.
		Also applies a linear interpolation on dataset (and raises warning).
		"""
		self.x, self.y_norm = savgol(self.x, self.y, self.ref, self.sam, window=window, order=order)
		self.y = self.y_norm
		self.ref = []
		self.sam = []
		warnings.warn('Linear interpolation have been applied to data.', InterpolationWarning)
		
	
	def slice(self, start=None, stop=None):
		"""
		Cuts the dataset on x axis in this form: [start, stop]

		Parameters:
		----------
		start: `float`
			start value of cutting interval
			Not giving a value will keep the dataset's original minimum value.
			Note that giving `None` will leave original minimum untouched too.
			Default is `None`.

		stop: `float`
			stop value of cutting interval
			Not giving a value will keep the dataset's original maximum value.
			Note that giving `None` will leave original maximum untouched too.
			Default is `None`.

		Notes:
		------

		If arms were given, it will merge them into the `self.y` and `self.y_norm` variables.
		"""
		self.x, self.y_norm = cut_data(self.x, self.y, self.ref, self.sam, start=start, stop=stop)
		self.ref = []
		self.sam = []
		self.y = self.y_norm
		# Just to make sure it's correctly shaped. Later on we might delete this.
		if type(self).__name__ == 'FFTMethod':
			self.original_x = self.x
		self._is_normalized = self._ensure_norm()

	def convolution(self, window_length, std=20):
		"""
		Applies a convolution with a gaussian on the dataset

		Parameters:
		----------
		window_length: `int`
			Length of the gaussian window.

		std: `float`
			Standard deviation of the gaussian
			Default is `20`.

		Notes:
		------
		If arms were given, it will merge them into the `self.y` and `self.y_norm` variables.
		Also applies a linear interpolation on dataset (and raises warning).
		"""
		self.x, self.y_norm = convolution(self.x, self.y, self.ref, self.sam, window_length, standev=std)
		self.ref = []
		self.sam = []
		self.y = self.y_norm
		warnings.warn('Linear interpolation have been applied to data.', InterpolationWarning)


	def detect_peak(self, pmax=0.1, pmin=0.1, threshold=0.1, except_around=None):
		"""
		Basic algorithm to find extremal points in data using ``scipy.signal.find_peaks``.

		Parameters:
		----------

		pmax: `float`
			Prominence of maximum points.
			The lower it is, the more peaks will be found.
			Default is `0.1`.

		pmin: float
			Prominence of minimum points.
			The lower it is, the more peaks will be found.
			Default is `0.1`.

		threshold: `float`
			Sets the minimum distance (measured on y axis) required for a point to be
			accepted as extremal.
			Default is 0.

		except_around: interval (array or tuple), 
			Overwrites the threshold to be 0 at the given interval.
			format is `(lower, higher)` or `[lower, higher]`.
			Default is None.

		Returns:
		-------
		xmax: `array-like`
			x coordinates of the maximums

		ymax: `array-like`
			y coordinates of the maximums

		xmin: `array-like`
			x coordinates of the minimums

		ymin: `array-like`
			y coordinates of the minimums
		"""
		x, y, ref, sam = self._safe_cast()
		xmax, ymax, xmin, ymin = find_peak(x, y, ref, sam, proMax=pmax, proMin=pmin, threshold=threshold, except_around=except_around)
		return xmax, ymax, xmin, ymin
		
	def show(self):
		"""
		Draws a graph of the current dataset using matplotlib.
		"""
		if np.iscomplexobj(self.y):
			self.plotwidget.plot(self.x, np.abs(self.y))
		else:   
			try:
				self.plotwidget.plot(self.x, self.y_norm, 'r')
			except Exception:
				self.plotwidget.plot(self.x, self.y, 'r')
		self.plotwidget.grid()
		self.plotwidget.show()


	def normalize(self, filename=None, smoothing_level=0):
		'''
		Normalize the interferogram by finding upper and lower envelope
		on an interactive matplotlib editor.

		Parameters
		----------

		filename: `str`
			Save the normalized interferogram named by filename in the 
			working directory. If not given it will not be saved.
			Default None.

		Returns
		-------
		None
		'''
		if run_from_ipython():
			return '''It seems you run this code in IPython. Interactive plotting is not yet supported. Consider running it in the regular console.'''
		x, y, _, _ = self._safe_cast()
		if smoothing_level != 0:
			x, y = savgol(x, y, [], [], window=smoothing_level)
		_l_env = DraggableEnvelope(x, y, 'l')
		y_transform = _l_env.get_data()
		_u_env = DraggableEnvelope(x, y_transform, 'u')
		y_final = _u_env.get_data()
		self.y = y_final
		self.y_norm = y_final
		self._is_normalized = True
		self.plotwidget.title('Final')
		self.show()
		if filename:
			if not filename.endswith('.txt'):
				filename += '.txt'
			np.savetxt(filename, np.transpose([self.x, self.y]), delimiter=',')
			print(f'Successfully saved as {filename}')


	def open_SPP_panel(self):
		_spp = SPPEditor(self.x, self.y_norm)
		self.delay, self.positions = _spp.get_data()

	def emit(self):
		# validate if it's typed by hand..
		if not isinstance(self._positions, np.ndarray):
			self._positions = np.asarray(self.positions)
		if not isinstance(self.delay, np.ndarray):
			self.delay = np.ones_like(self.positions) * self.delay
		return self.delay, self.positions

	def set_SPP_data(self, delay, positions):
		if not isinstance(positions, np.ndarray):
			positions = np.asarray(positions)
		if not isinstance(delay, float):
			delay = float(delay)
		delay = np.ones_like(positions) * delay
		self._delay = delay
		self._positions = positions



'''
This file is not finished by any means.
'''
from math import factorial

import numpy as np
import matplotlib.pyplot as plt

class Phase:
    '''
    A class that represents a phase obtained from various
    methods.
    '''
    is_dispersion_array = False
    is_coeff = False

    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.poly = None


    def __call__(self, value):
        if self.poly:
            return self.poly.__call__(value)
        raise NotImplementedError


    @classmethod
    def from_disperion_array(cls, dispersion_array, x_range=(2, 4)):
        cls.is_dispersion_array = True
        x = np.linspace(*x_range, num=1000)
        coeffs = [i * factorial(i) for i in dispersion_array] # nem jó
        cls.poly = np.poly1d(coeffs[::-1])
        return cls(x, cls.poly(x))


    @classmethod
    def from_coeff(cls, GD, GDD=0, TOD=0, FOD=0, QOD=0, SOD=0, x_range=(2, 4)):
        x = np.linspace(*x_range, num=1000)
        cls.is_coeff = True
        cls.poly = np.poly1d([SOD, QOD, FOD, TOD, GDD, GD])
        return cls(x, cls.poly(x))

    def __str__(self):
        if self.poly is not None:
            return self.poly.__str__()

    def plot(self, ax=None, **kwargs):
        if ax is None:
            ax = plt
        if not self.is_dispersion_array or not self.is_coeff:
            ax.plot(self.x, self.y, **kwargs)
        else:
            ax.plot(self.x, self.poly(self.x), **kwargs)


    def fit(self, order, reference_point):
        if self.is_coeff or self.is_dispersion_array:
            pass

    @property
    def data(self):
        return self.x, self.y




import sys
import collections
import unittest
from unittest.mock import patch

import numpy as np
import pandas as pd

from pysprint import Dataset
from pysprint.utils.exceptions import DatasetError


class TestEvaluate(unittest.TestCase):

	def setUp(self):
		self.x = np.arange(1, 1000, 1)
		self.y = np.sin(self.x)

	def tearDown(self):
		pass

	def test_constructor(self):
		ifg = Dataset(self.x, self.y)
		np.testing.assert_array_equal(ifg.data['x'], self.x)
		np.testing.assert_array_equal(ifg.data['y'], self.y)

	def test_prediction(self):
		ifg = Dataset([150,200], [3,4])
		assert ifg.probably_wavelength == True
		ifg = Dataset([1,2], [3,4])
		assert ifg.probably_wavelength == False


	def test_dtypes(self):
		ifg = Dataset([15,4], [3,4])
		assert type(ifg.x) == np.ndarray
		assert type(ifg.y) == np.ndarray
		with self.assertRaises(DatasetError):
			ifg = Dataset(['das', 6541], [1,2])
		with self.assertRaises(DatasetError):
			ifg = Dataset([2, 6541], ['das',2])

	def test_safe_casting(self):
		ifg = Dataset([15, 4], [3, 4], [14,54], [45, 51])
		x, y, ref, sam = ifg._safe_cast()
		assert x is not ifg.x
		np.testing.assert_array_equal(x, ifg.x)
		assert y is not ifg.y
		np.testing.assert_array_equal(y, ifg.y)
		assert ref is not ifg.ref
		np.testing.assert_array_equal(ref, ifg.ref)
		assert sam is not ifg.sam
		np.testing.assert_array_equal(sam, ifg.sam)

	def test_rawparsing(self):
		ifg = Dataset.parse_raw('test_rawparsing.trt')
		assert issubclass(ifg.meta.__class__, collections.abc.Mapping)
		with self.assertRaises(OSError):
			ifg = Dataset.parse_raw(546)


	def test_data(self):
		ifg = Dataset(self.x, self.y)
		assert isinstance(ifg.data, pd.DataFrame)

	@patch('matplotlib.pyplot.show')
	def test_plotting(self, mock_show):
		ifg = Dataset(self.x, self.y)
		ifg.show()
		mock_show.assert_called()

	def test_chdomain(self):
		ifg = Dataset(self.x, self.y)
		before = ifg.x
		ifg.chdomain()
		ifg.chdomain()
		after = ifg.x
		np.testing.assert_array_almost_equal(before, after)

	# fails on azure pipelines.. should be fixed
	# @patch('matplotlib.pyplot.show')
	# def test_normalize(self, mock_show):
	# 	ifg = Dataset(self.x, self.y)
	# 	ifg.normalize()
	# 	mock_show.assert_called()

	@patch('matplotlib.pyplot.show')
	def test_sppeditor(self, mock_show):
		ifg = Dataset(self.x, self.y)
		ifg.open_SPP_panel()
		mock_show.assert_called()

	def test_spp_setting(self):
		ifg = Dataset(self.x, self.y)
		delay = 100
		positions = [200, 300]
		ifg.set_SPP_data(delay, positions)
		delay, position = ifg.emit()
		assert len(delay) == len(position)
		np.testing.assert_array_equal(delay, np.array([100, 100]))
		np.testing.assert_array_equal(position, positions)


if __name__ == '__main__':
	unittest.main()



import numpy as np

from pysprint.core.bases.dataset import Dataset
from pysprint.mpl_tools.peak import EditPeak
from pysprint.utils import _maybe_increase_before_cwt, print_disp, run_from_ipython, calc_envelope
from pysprint.core.evaluate import min_max_method

__all__ = ['MinMaxMethod']

class MinMaxMethod(Dataset):
	"""
	Basic interface for Minimum-Maximum Method.
	"""
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

	# TODO: fix docstring
	def init_edit_session(self, engine='normal', **kwargs):
		""" Function to initialize peak editing on a plot.
		Right clicks will delete the closest point, left clicks 
		will add a new point. Just close the window when finished.
		
		Parameters:
		----------

		engine: str, default is `'normal'`
			Must be `'cwt'`, `'normal'` or `'slope'`.
			Peak detection algorithm to use.
			
		**kwargs:
			pmax, pmin, threshold, except_around, width

		Notes:
		------

		Currently this function is disabled when running it from IPython.
		"""
		if run_from_ipython():
			return '''It seems you run this code in IPython. Interactive plotting is not yet supported. Consider running it in the regular console.'''
		engines = ('cwt', 'normal', 'slope')
		if engine not in engines:
			raise ValueError(f'Engine must be in {str(engines)}')
		if engine == 'normal':
			pmax = kwargs.pop('pmax', 0.1)
			pmin = kwargs.pop('pmin', 0.1)
			threshold = kwargs.pop('threshold', 0.1)
			except_around = kwargs.pop('except_around', None)
			_x, _y, _xx, _yy = self.detect_peak(
			pmax=pmax, pmin=pmin, threshold=threshold, except_around=except_around
			)
			
			# just for validation purposes
			_ = kwargs.pop('width', 10)
			_ = kwargs.pop('floor_thres', 0.05)

		elif engine == 'slope':
			x, _, _, _ = self._safe_cast()
			y = np.copy(self.y_norm)
			if _maybe_increase_before_cwt(y):
				y += 2
			_, lp, lloc = calc_envelope(y, np.arange(len(y)), 'l')
			_, up, uloc = calc_envelope(y, np.arange(len(y)), 'u')
			lp -= 2
			up -= 2
			_x, _xx = x[lloc], x[uloc]
			_y, _yy = lp, up

		elif engine == 'cwt':
			width = kwargs.pop('width', 10)
			floor_thres = kwargs.pop('floor_thres', 0.05)
			_x, _y, _xx, _yy = self.detect_peak_cwt(width=width, floor_thres=floor_thres)

			# just for validation purposes
			_ = kwargs.pop('pmax', 0.1)
			_ = kwargs.pop('pmin', 0.1)
			_ = kwargs.pop('threshold', 0.1)
			_ = kwargs.pop('except_around', None)

		_xm = np.append(_x, _xx)
		_ym = np.append(_y, _yy)


		if kwargs:
			raise TypeError(f'Invalid argument:{kwargs}')

		try:
			_editpeak = EditPeak(self.x, self.y_norm, _xm, _ym)
		except ValueError:
			_editpeak = EditPeak(self.x, self.y, _xm, _ym)
		# automatically propagate these points to the mins and maxes
		# better distribute these points between min and max, just in case
		# the default argrelextrema is definitely not called
		# in `pysprint.core.evaluate.min_max_method`.
		self.xmin = _editpeak.get_dat[0][:len(_editpeak.get_dat[0])//2]
		self.xmax = _editpeak.get_dat[0][len(_editpeak.get_dat[0])//2:] 
		print(f'In total {len(_editpeak.get_dat[0])} extremal points were recorded.')
		return _editpeak.get_dat[0]

	@print_disp
	def calculate(self, reference_point, order, show_graph=False):
		""" 
		MinMaxMethod's calculate function.

		Parameters:
		----------

		reference_point: float
			reference point on x axis

		fit_order: int
			Polynomial (and maximum dispersion) order to fit. Must be in [1,5].

		show_graph: bool
			shows a the final graph of the spectral phase and fitted curve.

		Returns:
		-------

		dispersion: array-like
			[GD, GDD, TOD, FOD, QOD]

		dispersion_std: array-like
			standard deviations due to uncertanity of the fit
			[GD_std, GDD_std, TOD_std, FOD_std, QOD_std]

		fit_report: lmfit report
			if lmfit is available, the fit report


		Notes:
		------

		Decorated with print_disp, so the results are immediately printed without explicitly saying so.
		"""
		dispersion, dispersion_std, fit_report = min_max_method(
			self.x, self.y, self.ref, self.sam, ref_point=reference_point,
			maxx=self.xmax, minx=self.xmin, fit_order=order, show_graph=show_graph
			)
		self._dispersion_array = dispersion
		return dispersion, dispersion_std, fit_report


import numpy as np


__all__ = [
    'cos_fit1', 'cos_fit2', 'cos_fit3', 'cos_fit4', 'cos_fit5', 'cos_fit6',
    'poly1', 'poly2', 'poly3', 'poly4','poly5', 'poly6'
    ]


def poly6(x, b0, b1, b2, b3, b4, b5, b6):
    """
    Taylor polynomial for fit
    b1 = GD
    b2 = GDD / 2
    b3 = TOD / 6
    b4 = FOD / 24
    b5 = QOD / 120
    b6 = SOD / 720
    """
    return b0+b1*x+b2*x**2+b3*x**3+b4*x**4+b5*x**5+b6*x**6


def poly5(x, b0, b1, b2, b3, b4, b5):
    """
    Taylor polynomial for fit
    b1 = GD
    b2 = GDD / 2
    b3 = TOD / 6
    b4 = FOD / 24
    b5 = QOD / 120
    """
    return b0+b1*x+b2*x**2+b3*x**3+b4*x**4+b5*x**5


def poly4(x, b0, b1, b2, b3, b4):
    """
    Taylor polynomial for fit
    b1 = GD
    b2 = GDD / 2
    b3 = TOD / 6
    b4 = FOD / 24
    """
    return b0+b1*x+b2*x**2+b3*x**3+b4*x**4


def poly3(x, b0, b1, b2, b3):
    """
    Taylor polynomial for fit
    b1 = GD
    b2 = GDD / 2
    b3 = TOD / 6

    """
    return b0+b1*x+b2*x**2+b3*x**3


def poly2(x, b0, b1, b2):
    """
    Taylor polynomial for fit
    b1 = GD
    b2 = GDD / 2
    """
    return b0+b1*x+b2*x**2


def poly1(x, b0, b1):
    """
    Taylor polynomial for fit
    b1 = GD
    """
    return b0+b1*x


def cos_fit1(x, c0, c1, b0, b1):
    return c0 + c1 * np.cos(poly1(x, b0, b1))


def cos_fit2(x, c0, c1, b0, b1, b2):
    return c0 + c1 * np.cos(poly2(x, b0, b1, b2))


def cos_fit3(x, c0, c1, b0, b1, b2, b3):
    return c0 + c1 * np.cos(poly3(x, b0, b1, b2, b3))


def cos_fit4(x, c0, c1, b0, b1, b2, b3, b4):
    return c0 + c1 * np.cos(poly4(x, b0, b1, b2, b3, b4))


def cos_fit5(x, c0, c1, b0, b1, b2, b3, b4, b5):
    return c0 + c1 * np.cos(poly5(x, b0, b1, b2, b3, b4, b5))


def cos_fit6(x, c0, c1, b0, b1, b2, b3, b4, b5, b6):
    return c0 + c1 * np.cos(poly6(x, b0, b1, b2, b3, b4, b5, b6))


import unittest
from unittest.mock import patch

import numpy as np

import pysprint
from pysprint.core.evaluate import spp_method
from pysprint import SPPMethod


class TestEvaluate(unittest.TestCase):

    def setUp(self):

        self.delays = np.array([-1700.,-1700.,-1500., -1500., -1300., -1300., -1100., -1100.,  -900.,  -900.,
        -700.,  -700.,  -500.,  -500.,  -300.,  -300.,  -100.,  -100.,   100.,   100.,
        300.])

        self.delays_invalid = np.array([-1700.,-1700.,-1500., -1500., -1300., -1300., -1100., -1100.,  -900.,  -900.,
          -700.,  -700.,  -500., np.nan,  -500.,  -300.,  -300.,  -100.,  -100.,   100.,   100.,
           300., 900])

        self.omegas = np.array([2.15686148, 2.55201208, 2.16708089, 2.54264428, 2.17900354, 2.53072164,
        2.19092619, 2.51965061, 2.20284884, 2.50772796, 2.21732634, 2.49325046,
        2.23180384, 2.47536649, 2.24968781, 2.45918575, 2.27097825, 2.43874693,
        2.29993325, 2.40894031, 2.33825604])

    def test_spp_in_core(self):
        x, y, d, ds, _ = spp_method(self.delays, self.omegas, fit_order=2, ref_point=2.355)
        np.testing.assert_array_equal(d, [258.84297727172856, -21.572879102888976, -100426.4054547129, 0, 0])

    def test_spp_exceptions(self):
        with self.assertRaises(ValueError):
            x, y, d, ds, _ = spp_method(self.delays, self.omegas, ref_point=0, fit_order=8)
        with self.assertRaises(ValueError):
            x, y, d, ds, _ = spp_method(self.delays, self.omegas, ref_point=0, fit_order=3.5)

    def test_spp_invalid(self):
        with self.assertRaises(ValueError):
            x, y, d, ds, _ = spp_method(self.delays_invalid, self.omegas, fit_order=2, ref_point=2.355)

    def test_spp_not_enough_vars(self):
        with self.assertRaises(TypeError):
            x, y, d, ds, _ = spp_method(np.array([1, 2, 3]), np.array([4, 5, 6]), fit_order=4, ref_point=2.355)

    def test_spp_from_raw_api(self):
        with patch('matplotlib.pyplot.show') as mock_show:
            ifgs = SPPMethod('')
            ifgs.listen(self.delays, self.omegas)
            d, ds, _ = ifgs.calculate(2.355, 2, show_graph=True)
            np.testing.assert_array_equal(d, [258.84297727172856, -21.572879102888976, -100426.4054547129])
            mock_show.assert_called()


if __name__ == '__main__':
    unittest.main()


'''
Experimental
'''
import re

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.widgets import TextBox


class SPPEditor:

    epsilon = 5  # max absolute pixel distance to count as a hit

    def __init__(self, x, y):
        self.fig, self.ax = plt.subplots()
        self.fig.set_figheight(6)
        self.fig.set_figwidth(10)
        plt.grid()
        plt.subplots_adjust(bottom=0.2)
        self.x = x
        self.y = y
        self._ind = None  # the active point index
        self.x_pos, self.y_pos = np.array([]), np.array([])
        self.basedata, = self.ax.plot(self.x, self.y)
        self.points, = self.ax.plot(self.x_pos, self.y_pos, 'ko')
        self.fig.canvas.mpl_connect(
            'key_press_event', self.key_press_callback
            )
        self.fig.canvas.mpl_connect(
            'button_release_event', self.button_release_callback
            )
        self.axbox = plt.axes([0.1, 0.05, 0.8, 0.1])
        self.text_box = TextBox(self.axbox, 'Delay [fs]', initial='0')
        self.text_box.on_submit(self.submit)
        self.text_box.on_text_change(self.text_change)
        plt.show()

    def submit(self, delay):
        try:
            delay = re.sub(r'[^0-9\.,\-]', "", delay)
            self.delay = float(delay)
        except ValueError:
            pass  # we ignore bad calls

    def text_change(self, delay):
        try:
            delay = re.sub(r'[^0-9\.,\-]', "", delay)
            self.delay = float(delay)
        except ValueError:
            pass  # we ignore bad calls

    def get_data(self):
        positions, _ = self.points.get_data()
        if not hasattr(self, 'delay'):
            self.delay = 0
        self.delay = np.ones_like(positions) * self.delay
        return self.delay, positions

    def button_release_callback(self, event):
        '''whenever a mouse button is released'''
        if event.button != 1:
            return
        self._ind = None

    def get_ind_under_point(self, event):
        '''
        Get the index of the selected point within the given epsilon tolerance
        '''

        # We use the pixel coordinates, because the axes are usually really
        # differently scaled.
        if event.inaxes is None:
            return
        if event.inaxes in [self.ax]:
            try:
                x, y = self.points.get_data()
                xy_pixels = self.ax.transData.transform(np.vstack([x, y]).T)
                xpix, ypix = xy_pixels.T

                # return the index of the point iff within epsilon distance.
                d = np.hypot(xpix - event.x, ypix - event.y)
                indseq, = np.nonzero(d == d.min())
                ind = indseq[0]

                if d[ind] >= self.epsilon:
                    ind = None
            except ValueError:
                return
            return ind

    def key_press_callback(self, event):
        '''whenever a key is pressed'''
        if not event.inaxes:
            return
        if event.key == 'd':
            if event.inaxes in [self.ax]:
                ind = self.get_ind_under_point(event)
            else:
                ind = None
            if ind is not None:
                self.x_pos = np.delete(self.x_pos, ind)
                self.y_pos = np.delete(self.y_pos, ind)
                self.points.set_data(self.x_pos, self.y_pos)

        elif event.key == 'i':
            if event.inaxes in [self.ax]:
                self.x_pos = np.append(self.x_pos, event.xdata)
                self.y_pos = np.append(self.y_pos, event.ydata)
                self.points.set_data(self.x_pos, self.y_pos)

        if self.points.stale:
            self.fig.canvas.draw_idle()



from math import factorial

import numpy as np

from pysprint.core.bases.dataset import Dataset
from pysprint.core.optimizer import FitOptimizer
from pysprint.utils import print_disp
from pysprint.core.evaluate import cff_method


__all__ = ['CosFitMethod']


class CosFitMethod(Dataset):
	"""
	Basic interface for the Cosine Function Fit Method.
	"""
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.params = [1, 1, 1, 1, 1, 1, 1, 1]
		self.fit = None
		self.mt = 8000
		self.f = None


	def set_max_tries(self, value):
		"""
		Overwrite the default scipy maximum try setting to fit the curve.
		"""
		self.mt = value

	def adjust_offset(self, value):
		"""
		Initial guess for offset
		"""
		self.params[0] = value

	def adjust_amplitude(self, value):
		"""
		Initial guess for amplitude
		"""
		self.params[1] = value

	def guess_GD(self, value):
		"""
		Initial guess for GD in fs
		"""
		self.params[3] = value

	def guess_GDD(self, value):
		"""
		Initial guess for GDD in fs^2
		"""
		self.params[4] = value / 2

	def guess_TOD(self, value):
		"""
		Initial guess for TOD in fs^3
		"""
		self.params[5] = value / 6

	def guess_FOD(self, value):
		"""
		Initial guess for FOD in fs^4
		"""
		self.params[6] = value / 24

	def guess_QOD(self, value):
		"""
		Initial guess for QOD in fs^5
		"""
		self.params[7] = value / 120

	def set_max_order(self, order):
		"""
		Sets the maximum order of dispersion to look for.
		Should be called after guessing the initial parameters.

		Parameters:
		----------

		order: int
			maximum order of dispersion to look for. Must be in [1, 5]
		"""
		if order > 5 or order < 1:
			print(f'Order should be an in integer from [1,5], currently {order} is given')
		try:
			int(order)
		except ValueError:
			print(f'Order should be an in integer from [1,5], currently {order} is given')
		order = 6 - order
		for i in range(1, order):
			self.params[-i] = 0

	@print_disp
	def calculate(self, reference_point):
		""" 
		Cosine fit's calculate function.

		Parameters:
		----------
		reference_point: float
			reference point on x axis

		Returns:
		-------

		dispersion: array-like
			[GD, GDD, TOD, FOD, QOD]

		dispersion_std: array-like
			standard deviations due to uncertanity of the fit
			[GD_std, GDD_std, TOD_std, FOD_std, QOD_std]

		fit_report: lmfit report
			WILL BE IMPLEMENTED


		Notes:
		------

		Decorated with print_disp, so the results are immediately printed without explicitly saying so.
		"""
		dispersion, self.fit = cff_method(self.x, self.y, self.ref, self.sam, 
			ref_point=reference_point, p0=self.params, maxtries=self.mt)
		dispersion = list(dispersion)
		while len(dispersion) < 5:
			dispersion.append(0)
		return dispersion, [0, 0, 0, 0, 0], 'Fit report for CFF not supported yet.'

	def plot_result(self):
		"""
		If the fitting happened, draws the fitted curve on the original dataset.
		Also prints the coeffitient of determination of the fit (a.k.a. r^2).
		"""
		try:
			residuals = self.y_norm - self.fit
			ss_res = np.sum(residuals**2)
			ss_tot = np.sum((self.y_norm - np.mean(self.y_norm))**2)
			print('r^2 = ' + str(1 - (ss_res / ss_tot)))
		except Exception:
			pass
		if self.fit is not None:
			self.plotwidget.plot(self.x, self.fit, 'k--', label='fit', zorder=99)
			self.plotwidget.legend()
			self.show()
		else:
			self.show()

	def optimizer(self, reference_point, order=3, initial_region_ratio=0.1,
		extend_by=0.1, coef_threshold=0.3, max_tries=5000, show_endpoint=True):
		"""
		Cosine fit optimizer. It's based on adding new terms to fit function successively
		until we reach the max_order.

		Notes
		-----
		If the fit fails some parameters must be tweaked in order to achieve results.
		There is a list below with issues, its suspected reasons and solutions.

		**SciPy raises OptimizeWarning and the affected area is small or not showing
		  any fit

		Reasons:
		- Completely wrong initial GD guess (or lack of guessing).
		- Too broad inital region, so that the optimizer cannot find a suitable fit.
		  This usually happens when the used data is large, or the spectral resolution
		  is high.

		Solution:
		- Provide better inital guess for GD.
		- Lower the inital_region_ratio.

		**SciPy raises OptimizeWarning and the affected area is bigger

		Reasons:
		- When the optimizer steps up with order it also extends the region of fit.
		This error usually present when the region of fit is too quickly growing.

		Solution:
		- Lower extend_by argument.

		**The optimizer is finished, but wrong fit is produced.

		Reasons:
		- We measure the goodness of fit with r^2 value. To allow this
		optimizer to smoothly find appropriate fits even for noisy datasets
		it's a good practice to keep the r^2 a lower value, such as the default 0.3.
		The way it works is we step up in order of fit (until max order) and extend
		region every time when a fit reaches the specified r^2 threshold value.
		This can be controlled via the coef_threshold argument.

		Solution:
		- Adjust the coef_threshold value. Note that it's highly recommended not to
		set a higher value than 0.6.
		"""

		x, y, ref, sam = self._safe_cast()
		self.f = FitOptimizer(x, y, ref, sam, reference_point=reference_point,
		max_order=order)
		self.f.set_initial_region(initial_region_ratio)
		self.f.set_final_guess(GD=self.params[3], GDD=self.params[4], TOD=self.params[5],
		FOD=self.params[6], QOD=self.params[7]) # we can pass it higher params safely, they are ignored.
		disp = self.f.run(extend_by, coef_threshold, max_tries=max_tries, show_endpoint=show_endpoint)
		disp = disp[3:]
		retval = [disp[i]*factorial(i+1) for i in range(len(disp))]
		self._dispersion_array = retval
		return retval



# -*- coding: utf-8 -*-
from math import factorial

import numpy as np
from scipy import fftpack
from scipy.optimize import curve_fit
from scipy.signal import argrelextrema

try:
    from lmfit import Model
    _has_lmfit = True
except ImportError:
    _has_lmfit = False


from pysprint.utils import (
    find_nearest, _handle_input, unpack_lmfit, fourier_interpolate,
    transform_cf_params_to_dispersion, transform_lmfit_params_to_dispersion,
    plot_phase
    )

from pysprint.core.functions import *


__all__ = [
    'min_max_method', 'spp_method', 'cff_method', 'fft_method',
    'cut_gaussian', 'ifft_method', 'args_comp'
    ]


_fit_config = {
    1: poly1,
    2: poly2,
    3: poly3,
    4: poly4,
    5: poly5
}

_cosfit_config = {
    1: cos_fit1,
    2: cos_fit2,
    3: cos_fit3,
    4: cos_fit4,
    5: cos_fit5
}


def min_max_method(
        x, y, ref, sam, ref_point,
        maxx=None, minx=None, fit_order=5, show_graph=False
        ):
    """Calculates the dispersion with minimum-maximum method

    Parameters
    ----------

    x: array-like
    x-axis data

    y: array-like
    y-axis data

    ref, sam: array-like
    reference and sample arm spectra evaluated at x

    ref_point: float
    the reference point to calculate order

    maxx and minx: array-like, optional
    the accepted minimal and maximal x values (if you want to manually pass)

    fit_order: int, optional
    degree of polynomial to fit data [1, 5]

    show_graph: bool, optional
    if True plot the calculated phase

    Returns
    -------

    dispersion: array-like
    [GD, GDD, TOD, FOD, QOD]

    dispersion_std: array-like
    [GD_std, GDD_std, TOD_std, FOD_std, QOD_std]

    fit_report: lmfit report
    """
    if fit_order not in range(6):
        raise ValueError('fit order must be in [1, 5]')

    x, y = _handle_input(x, y, ref, sam)
    if maxx is None:
        maxInd = argrelextrema(y, np.greater)
        maxx = x[maxInd]
    if minx is None:
        minInd = argrelextrema(y, np.less)
        minx = x[minInd]

    _, ref_index = find_nearest(x, ref_point)

    max_freq = x[ref_index] - maxx
    min_freq = x[ref_index] - minx

    neg_freq = np.sort(
        np.append(max_freq[max_freq < 0], min_freq[min_freq < 0])
        )[::-1]
    pos_freq = np.sort(
        np.append(max_freq[max_freq > 0], min_freq[min_freq > 0])
        )

    if len(neg_freq) == 0 and len(pos_freq) == 0:
        raise ValueError('No extremal points found.')

    pos_values = np.pi * np.arange(1, len(pos_freq) + 1)
    neg_values = np.pi * np.arange(1, len(neg_freq) + 1)

    x_s = np.append(pos_freq, neg_freq)
    y_s = np.append(pos_values, neg_values)

    # FIXME: Do we even need this?
    # Yes, we do. This generates a prettier plot.
    idx = np.argsort(x_s)
    full_x, full_y = x_s[idx], y_s[idx]

    _function = _fit_config[fit_order]

    if _has_lmfit:
        fitmodel = Model(_function)
        pars = fitmodel.make_params(
            **{f'b{i}': 1 for i in range(fit_order + 1)}
            )
        result = fitmodel.fit(full_y, x=full_x, params=pars)
    else:
        popt, pcov = curve_fit(_function, full_x, full_y, maxfev=8000)

    try:
        if _has_lmfit:
            dispersion, dispersion_std = transform_lmfit_params_to_dispersion(
                *unpack_lmfit(result.params.items()), drop_first=True, dof=1
                )
            fit_report = result.fit_report()
        else:
            dispersion, dispersion_std = transform_cf_params_to_dispersion(
                popt, drop_first=True
                )
            fit_report = ('To display detailed results,'
                          ' you must have `lmfit` installed.')
        if show_graph:
            plot_phase(full_x, full_y, bf=result.best_fit,
                       bf_fallback=_function(full_x, *popt),
                       window_title='Min-max method fitted')
        return dispersion, dispersion_std, fit_report
    except Exception as e:
        raise e


def spp_method(delays, omegas, ref_point=0, fit_order=4):
    """
    Calculates the dispersion from SPP's positions and delays.

    Parameters
    ----------

    delays: array-like
    The delay values in fs exactly where omegas taken.

    omegas: array-like
    The angular frequency values where SPP's are located

    fit_order: int
    order of polynomial to fit the given data

    Returns
    -------
    omegas: array-like
    x axis data

    delays: array-like
    y axis data

    dispersion: array-like
    [GD, GDD, TOD, FOD, QOD]

    dispersion_std: array-like
    [GD_std, GDD_std, TOD_std, FOD_std, QOD_std]

    bf: array-like
    best fitting curve for plotting
    """
    if fit_order not in range(5):
        raise ValueError('fit order must be in [1, 4]')

    omegas = np.asarray(omegas).astype(np.float64)

    delays = np.asarray(delays).astype(np.float64)

    if not len(delays) == len(omegas):
        raise ValueError(f'data shapes are different: {delays.shape} & {omegas.shape}')

    idx = np.argsort(omegas)
    omegas, delays = omegas[idx], delays[idx]
    omegas -= ref_point

    _function = _fit_config[fit_order]

    try:
        if _has_lmfit:
            fitmodel = Model(_function)
            pars = fitmodel.make_params(
                **{f'b{i}': 1 for i in range(fit_order + 1)}
                )
            result = fitmodel.fit(delays, x=omegas, params=pars)

            dispersion, dispersion_std = transform_lmfit_params_to_dispersion(
                *unpack_lmfit(result.params.items()), drop_first=False, dof=0
                )
            bf = result.best_fit
        else:
            popt, pcov = curve_fit(_function, omegas, delays, maxfev=8000)
            dispersion, dispersion_std = transform_cf_params_to_dispersion(
                popt, drop_first=False
                )
            bf = _function(omegas, *popt)
        return omegas, delays, dispersion, dispersion_std, bf

    except Exception as e:
        raise e # this should be deleted..


def cff_method(
        x, y, ref, sam, ref_point=0,
        p0=[1, 1, 1, 1, 1, 1, 1, 1], maxtries=8000
        ):
    """
    Phase modulated cosine function fit method.

    Parameters
    ----------

    x: array-like
    x-axis data

    y: array-like
    y-axis data

    ref, sam: array-like
    the reference and sample arm spectra evaluated at x

    p0: array-like
    the initial parameters for fitting

    Returns
    -------

    dispersion: array-like
    [GD, GDD, TOD, FOD, QOD]

    bf: array-like
    best fitting curve

    """

    x, y = _handle_input(x, y, ref, sam)

    try:
        orderhelper = np.max(np.flatnonzero(p0)) - 2

        p0 = np.trim_zeros(p0, 'b')

        _funct = _cosfit_config[orderhelper]

        popt, pcov = curve_fit(
            _funct, x - ref_point, y, p0, maxfev=maxtries
            )

        dispersion = np.zeros_like(popt)[:-3]
        for num in range(len(popt) - 3):
            dispersion[num] = popt[num + 3] * factorial(num + 1)
        return dispersion, _funct(x - ref_point, *popt)
    except RuntimeError:
        raise ValueError(f'''Max tries ({maxtries}) reached..
                             Parameters could not be estimated.''')


def fft_method(x, y):
    """Perfoms FFT on data

    Parameters
    ----------

    x: array-like
    the x-axis data

    y: array-like
    the y-axis data

    Returns
    -------
    xf: array-like
    the transformed x data

    yf: array-like
    transformed y data
    """
    yf = fftpack.fft(y)
    xf = np.linspace(x[0], x[-1], len(x))
    return xf, yf


def gaussian_window(t, tau, fwhm, order):
    """
    Returns a simple gaussian window of given parameters evaulated at t.

    Parameters
    ----------
    t: array-like
    input array to perform window on

    tau: float
    center of gaussian window

    fwhm: float
    FWHM of given gaussian

    order: float
    order of gaussian window. If not even it's incremented by 1.

    Returns
    -------
    array : array-like
    nth order gaussian window with params above

    """
    if order % 2 == 1:
        order += 1
    std = fwhm/(2 * (np.log(2)**(1 / order)))
    return np.exp(-((t - tau)**order)/(std**order))


def cut_gaussian(x, y, spike, fwhm, win_order):
    """
    Applies gaussian window with the given params.

    Parameters
    ----------
    x: array-like
    x-axis data

    y: array-like
    y-axis data

    spike: float
    center of gaussian window

    fwhm: float
    Full width at half max

    Returns
    -------

    y: array-like
    the windowed y values
    """
    y *= gaussian_window(x, tau=spike, fwhm=fwhm, order=win_order)
    return y


def ifft_method(x, y, interpolate=True):
    """
    Perfoms IFFT on data

    Parameters
    ----------

    x: array-like
    the x-axis data

    y: array-like
    the y-axis data

    interpolate: bool
    if True perform a linear interpolation on dataset before transforming

    Returns
    -------
    xf: array-like
    the transformed x data

    yf: array-like
    transformed y data

    """
    N = len(x)
    if interpolate:
        x, y = fourier_interpolate(x, y)
    xf = np.fft.fftfreq(N, d=(x[1]-x[0])/(2*np.pi))
    yf = np.fft.ifft(y)
    return xf, yf


def args_comp(x, y, ref_point=0, fit_order=5, show_graph=False):
    """
    Calculates the phase of complex dataset then unwrap by changing
    deltas between values to 2*pi complement. At the end, fit a
    polynomial curve to determine dispersion coeffs.

    Parameters
    ----------

    x: array-like
    the x-axis data

    y: array-like
    the y-axis data

    ref_point: float
    the reference point to calculate order

    fit_order: int
    degree of polynomial to fit data [1, 5]

    show_graph: bool
    if True return a plot with the results

    Returns
    -------

    dispersion: array-like
    [GD, GDD, TOD, FOD, QOD]

    dispersion_std: array-like
    [GD_std, GDD_std, TOD_std, FOD_std, QOD_std]

    fit_report: lmfit report
    """
    if fit_order not in range(6):
        raise ValueError('fit order must be in [1, 5]')

    x -= ref_point
    y = np.unwrap(np.angle(y), axis=0)

    _function = _fit_config[fit_order]



    if _has_lmfit:
        fitmodel = Model(_function)
        pars = fitmodel.make_params(
            **{f'b{i}': 1 for i in range(fit_order + 1)}
            )
        result = fitmodel.fit(y, x=x, params=pars)
    else:
        popt, pcov = curve_fit(_function, x, y, maxfev=8000)

    try:
        if _has_lmfit:
            dispersion, dispersion_std = transform_lmfit_params_to_dispersion(
                *unpack_lmfit(result.params.items()), drop_first=True, dof=1
                )
            fit_report = result.fit_report()
        else:
            dispersion, dispersion_std = transform_cf_params_to_dispersion(
                popt, drop_first=True
                )
            fit_report = ('To display detailed results,'
                          ' you must have `lmfit` installed.')
        if show_graph:
            plot_phase(x, y, result.best_fit, window_title='Phase')
        return -dispersion, dispersion_std, fit_report
    except Exception as e:
        raise e



import numpy as np
import matplotlib.pyplot as plt

from pysprint.core.bases.dataset_base import DatasetBase, C_LIGHT
from pysprint.core._generator import generatorFreq, generatorWave

__all__ = ['Generator']

class Generator(DatasetBase):
	"""
	Basic dataset generator.
	"""
	def __init__(self, start, stop, center, delay=0,
		GD=0, GDD=0, TOD=0, FOD=0, QOD=0, resolution=0.1,
	 	delimiter=',', pulse_width=10, normalize=False, chirp=0):
		self.start = start
		self.stop = stop
		self.center = center
		self.delay = delay
		self.GD = GD
		self.GDD = GDD
		self.TOD = TOD
		self.FOD = FOD
		self.QOD = QOD
		self.resolution = resolution
		self.delimiter = delimiter
		self.pulse_width = pulse_width
		self.chirp = chirp
		self.normalize = normalize
		self.x = np.array([])
		self.y = np.array([])
		self.ref = np.array([])
		self.sam = np.array([])
		self.plotwidget = plt
		self.is_wave = False
		
	def __str__(self):
		return self.__repr__()

	def __repr__(self):
		return f'''Generator({self.start}, {self.stop}, {self.center}, delay = {self.delay}, GD={self.GD}, GDD={self.GDD}, TOD={self.TOD}, FOD={self.FOD}, QOD={self.QOD}, resolution={self.resolution}, delimiter={self.delimiter}, pulse_width={self.pulse_width}, normalize={self.normalize})'''

	def _check_norm(self):
		"""
		Do the normalization when we can.
		"""
		if len(self.ref) != 0:
			self._y =  (self.y - self.ref - self.sam)/(2*np.sqrt(self.sam*self.ref))

	def generate_freq(self):
		"""
		Use this to generate the spectrogram in ang. frequency domain.
		"""
		self.x, self.y, self.ref, self.sam = generatorFreq(self.start, self.stop, self.center, self.delay, self.GD,
			self.GDD, self.TOD, self.FOD, self.QOD,
			self.resolution, self.delimiter, self.pulse_width, self.normalize, self.chirp)

	def generate_wave(self):
		"""
		Use this to generate the spectrogram in wavelength domain.
		"""
		self.is_wave = True
		self.x, self.y, self.ref, self.sam = generatorWave(self.start, self.stop, self.center, self.delay, self.GD,
			self.GDD, self.TOD, self.FOD, self.QOD,
			self.resolution, self.delimiter, self.pulse_width, self.normalize, self.chirp)


	def GD_lookup(self, *args, **kwargs):
		return self.GD + self.delay

	def show(self):
		"""
		Draws the plot of the generated data.
		"""
		self._check_norm()
		if np.iscomplexobj(self.y):
			self.plotwidget.plot(self.x, np.abs(self.y))
		else:   
			try:
				self.plotwidget.plot(self.x, self._y, 'r')
			except Exception:
				self.plotwidget.plot(self.x, self.y, 'r')
		self.plotwidget.grid()
		self.plotwidget.show()

	def save(self, name, path=None):
		"""
		Saves the generated dataset with numpy.savetxt.

		Parameters:
		----------

		name: string
			Name of the output file. You shouldn't include the .txt at the end.

		path: string, default is None
			You can also specify the save path.
			e.g path='C:/examplefolder'
		"""
		if path is None:
			np.savetxt(f'{name}.txt', np.transpose([self.x, self.y, self.ref, self.sam]), delimiter=self.delimiter)
			print(f'Successfully saved as {name}')
		else:
			np.savetxt(
				f'{path}/{name}.txt',
				 np.transpose([self.x, self.y, self.ref, self.sam]),
				 delimiter = self.delimiter
				 )
			print(f'Successfully saved as {name}')

	def _phase(self, j):
		if self.is_wave:
			lam = np.arange(self.start, self.stop + self.resolution, self.resolution) 
			omega = (2 * np.pi * C_LIGHT) / lam 
			omega0 = (2 * np.pi * C_LIGHT) / self.center 
			j = omega - omega0
		else:
			lamend = (2 * np.pi * C_LIGHT) / self.start
			lamstart = (2 * np.pi * C_LIGHT) / self.stop
			lam = np.arange(lamstart, lamend + self.resolution, self.resolution)
			omega = (2 * np.pi * C_LIGHT) / lam 
			j = omega - self.center
		return (j + self.delay * j + j * self.GD + (self.GDD / 2) * j ** 2 
			   + (self.TOD / 6)* j ** 3 + (self.FOD / 24) * j ** 4 + (self.QOD / 120) * j ** 5)

	def phase_graph(self):
		"""
		Plots the spectrogram along with the spectral phase.
		"""
		self._check_norm()
		self.fig, self.ax = self.plotwidget.subplots(2, 1, figsize=(8, 7))
		self.plotwidget.subplots_adjust(top=0.95)
		self.fig.canvas.set_window_title('Spectrum and phase')
		try:
			self.ax[0].plot(self.x, self._y, 'r')
		except Exception:
			self.ax[0].plot(self.x, self.y, 'r')
		try:
			self.ax[1].plot(self.x, self._phase(self.x))
		except Exception:
			raise ValueError('''The spectrum is not generated yet.
			Use self.generate_freq() on frequency domain or self.generate_wave() on wavelength domain.''')
		self.ax[0].set(xlabel="Frequency/Wavelength", ylabel="Intensity")
		self.ax[1].set(xlabel="Frequency/Wavelength", ylabel="$\Phi $[rad]")
		self.ax[0].grid()
		self.ax[1].grid()
		self.plotwidget.show()

	@property
	def data(self):
		"""
		Unpacks the generated data.
		If arms are given it returns x, y, reference_y, sample_y
		Else returns x, y
		"""
		if len(self.ref) == 0:
			return self.x, self.y
		return self.x, self.y, self.ref, self.sam


import numpy as np

__all__ = ['correct_sign']

def correct_sign(x, flip_increasing=True):
	if not isinstance(x, np.ndarray):
		x = np.asarray(x)
	increasing_idx = np.where(np.diff(x) > 0)
	#inc_len = len(increasing_idx[0])
	decreasing_idx = np.where(np.diff(x) < 0)
	#dec_len = len(decreasing_idx[0])
	#stationary_idx = np.where(np.diff(x) == 0)
	increasing = group_consecutives(increasing_idx)
	decreasing = group_consecutives(decreasing_idx)
	if not (len(increasing) == 1 and len(decreasing) == 1):
			raise ValueError('Values could not be split into two strictly monotonic parts safely.')
	if flip_increasing:
		x[increasing_idx[0][0]] *= -1
		x[increasing_idx[0]+1] *= -1 
		return x
	x[decreasing_idx[0]] *= -1
	return x

def group_consecutives(arr, step=1):
	return np.split(arr, np.where(np.diff(arr) != step)[0] + 1)



import os

import numpy as np
import matplotlib.pyplot as plt

from pysprint.core.bases.dataset import Dataset
from pysprint.core.evaluate import spp_method
from pysprint.utils import print_disp
from pysprint.utils.exceptions import DatasetError


__all__ = ['SPPMethod']

class SPPMethod:
    def __init__(self, ifg_names, sam_names=None, ref_names=None, **kwargs):
        self.ifg_names = ifg_names
        if sam_names:
            self.sam_names = sam_names
        else:
            self.sam_names = None
        if ref_names:
            self.ref_names = ref_names
        else:
            self.ref_names = None
        self._validate()
        if self.sam_names:
            if not len(self.ifg_names) == len(self.sam_names):
                raise DatasetError('Missmatching length of files.')
        if self.ref_names:
            if not len(self.ifg_names) == len(self.ref_names):
                raise DatasetError('Missmatching length of files.')
        self.idx = 0
        self.skiprows = kwargs.pop('skiprows', 8)
        self.decimal = kwargs.pop('decimal', ',')
        self.sep = kwargs.pop('sep', ';')
        self.meta_len = kwargs.pop('meta_len', 4)

        if kwargs:
            raise TypeError(f'invalid keyword argument:{kwargs}')

        self._delay = {}
        self._positions = {}


    def __len__(self):
        return len(self.ifg_names)

    def __iter__(self):
        return self

    def __str__(self):
        return f'{type(self).__name__} object\nInterferogram count : {len(self)}'

    def __next__(self):
        if self.idx < len(self):
            try:
                d = Dataset.parse_raw(
                    self.ifg_names[self.idx], self.sam_names[self.idx], self.ref_names[self.idx],
                    skiprows=self.skiprows, decimal=self.decimal, sep=self.sep, meta_len=self.meta_len
                    )
            except TypeError:
                d = Dataset.parse_raw(
                    self.ifg_names[self.idx], skiprows=self.skiprows, decimal=self.decimal, 
                    sep=self.sep, meta_len=self.meta_len
                    )
            self.idx += 1
            return d
        raise StopIteration


    def __getitem__(self, key):
        try:
            dataframe = Dataset.parse_raw(
                self.ifg_names[key], self.sam_names[key], self.ref_names[key],
                skiprows=self.skiprows, decimal=self.decimal, sep=self.sep, meta_len=self.meta_len
                )
        except (TypeError, ValueError):
            dataframe = Dataset.parse_raw(
                self.ifg_names[key], skiprows=self.skiprows, decimal=self.decimal, sep=self.sep,
                 meta_len=self.meta_len
                 )
        return dataframe

    def _validate(self):
        for filename in self.ifg_names:
            if os.path.exists(filename):
                pass
            else:
                raise FileNotFoundError(f'''File named '{filename}' is not found.''')
        if self.sam_names:
            for sam in self.sam_names:
                if os.path.exists(sam):
                    pass
                else:
                    raise FileNotFoundError(f'''File named '{sam}' is not found.''')
        if self.ref_names:
            for ref in self.ref_names:
                if os.path.exists(ref):
                    pass
                else:
                    raise FileNotFoundError(f'''File named '{ref}' is not found.''')


    def listen(self, delay, position):
        self._delay[self.idx] = delay
        self._positions[self.idx] = position


    def save_data(self, filename):
        if not filename.endswith('.txt'):
            filename += '.txt'
        delay = np.concatenate([_ for _ in self._delay.values()]).ravel()
        position = np.concatenate([_ for _ in self._positions.values()]).ravel()
        np.savetxt(f'{filename}', np.transpose(np.array([position, delay])), delimiter=',')

    @staticmethod
    @print_disp
    def calculate_from_raw(omegas, delays, reference_point, order):
        x, y, dispersion, dispersion_std, bf = spp_method(
            delays, omegas, ref_point=reference_point, fit_order=order
            )
        return dispersion, dispersion_std, ''


    @print_disp
    def calculate(self, reference_point, order=2, show_graph=False):
        delays = np.concatenate([_ for _ in self._delay.values()]).ravel()
        positions = np.concatenate([_ for _ in self._positions.values()]).ravel()
        x, y, dispersion, dispersion_std, bf = spp_method(
            delays, positions, ref_point=reference_point, fit_order=order
            )
        if show_graph:
            plt.plot(x, y, 'o')
            try:
                plt.plot(x, bf, 'r--', zorder=1)
            except Exception as e:
                print(e)
            plt.grid()
            plt.show()
        return dispersion, dispersion_std, bf

    @property
    def info(self):
        self._info = f'Data is recorded from {len(self._delay)} interferograms ({len(self)} availabe in total)'
        return self._info



import sys
import unittest
from unittest.mock import patch

import numpy as np

from pysprint import MinMaxMethod



class TestEvaluate(unittest.TestCase):

	def setUp(self):
		self.x = np.arange(1, 1000, 1)
		self.y = np.sin(self.x)

	def tearDown(self):
		pass

	@patch('matplotlib.pyplot.show')
	def test_edit_session(self, mock_show):
		# ifg = MinMaxMethod(self.x, self.y)
		# ifg.init_edit_session(engine='cwt')
		# ifg.init_edit_session(engine='slope')
		# ifg.init_edit_session(engine='normal')
		# mock_show.assert_called()
		# with self.assertRaises(ValueError):
		# 	ifg.init_edit_session(engine='dfssdfasdf')
		pass

		# this does not work on Azure Pipelines.. should be fixed



if __name__ == '__main__':
	unittest.main()



import warnings

__all__ = ['PySprintWarning', 'DatasetError', 'InterpolationWarning', 'FourierWarning']


class PySprintWarning(Warning):
	'''
	Base pysprint warning class.
	'''
	pass


class InterpolationWarning(PySprintWarning):
	"""
	This warning is raised when a function applies linear interpolation on the data.
	"""
	pass


class FourierWarning(PySprintWarning):
	"""
	This warning is raised when FFT is called first instead of IFFT.
	Later on it will be improved. 
	For more details see help(pysprint.FFTMethod.calculate)
	"""
	pass


class DatasetError(Exception):
	"""
	This error is raised when invalid type of data encountered when initializing 
	a dataset or inherited object.
	"""
	pass



from abc import abstractmethod, ABCMeta

__all__ = ['DatasetBase']

C_LIGHT = 299.792458

class DatasetBase(metaclass=ABCMeta):
	"""Base metaclass that defines the interface
	for any interferogram."""

	@abstractmethod
	def __init__(self):
		pass

	@abstractmethod
	def GD_lookup(self, *args, **kwargs):
		''' Quick GD lookup'''

	@property
	@abstractmethod
	def data(self):
		'''The stored dataset.'''
		return None

	@abstractmethod
	def show(self):
		'''Plot the stored dataset. '''
		pass


import os
import contextlib
import warnings

import numpy as np
from scipy.fftpack import fftshift

from pysprint.core.bases.dataset import Dataset
from pysprint.core.ffts_non_uniform import nuifft
from pysprint.utils import print_disp
from pysprint.core.evaluate import (fft_method,	cut_gaussian, ifft_method,
	args_comp, gaussian_window)
from pysprint.utils.exceptions import *
from pysprint.core.ffts_auto import _run
from pysprint.core.phase import Phase

__all__ = ['FFTMethod']

class FFTMethod(Dataset):
	"""
	Basic interface for the Fourier transform method.
	# FIXME: bug when calling show_graph on calulcate method: it opens up a clean figure.
	"""
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		#  making sure it's not normalized
		if self._is_normalized:
			self.y_norm = self.y
			self._is_normalized = False
		self.original_x = self.x
		self.at = None
		self.std = None
		self.window_order = None
		self._ifft_called_first = False
		self.phase = None

	def shift(self, axis='x'):
		"""
		Equivalent to scipy.fftpack.fftshift, but it's easier to
		use this function instead, because we don't need to explicitly
		call the class' x and y attribute.
		
		Parameter(s):
		------------
		axis: str, default is 'x'
			either 'x', 'y' or 'both'
		"""
		if axis == 'x':
			self.x = fftshift(self.x)
		elif axis == 'y':
			self.y = fftshift(self.y)
		elif axis == 'both':
			self.y = fftshift(self.y)
			self.x = fftshift(self.x)
		else:
			raise ValueError(f'axis should be either x, y or both, currently {axis} is given.')

	def ifft(self, interpolate=True, usenifft=False, eps=1E-12, exponent='positive'):
		"""
		Applies ifft to the dataset.
		
		Parameter(s):
		------------

		interpolate: bool, default is True
			Whether to apply linear interpolation on the dataset before transforming.

		usenifft: bool, default is False
			Whether to use non unifrom fft

		...
		"""
		self._ifft_called_first = True
		if usenifft:
			x_spaced = np.linspace(self.x[0], self.x[-1], len(self.x))
			timestep = np.diff(x_spaced)[0]
			x_axis = np.fft.fftfreq(len(self.x), d=timestep/(2*np.pi))
			y_transform = nuifft(
				self.x, self.y, gl=len(self.x), df=(x_axis[1]-x_axis[0]), epsilon=eps, exponent=exponent
				)
			self.x, self.y = x_axis, np.fft.fftshift(y_transform)

		else:
			self.x, self.y = ifft_method(self.x, self.y, interpolate=interpolate)

	def fft(self):
		"""
		Applies fft to the dataset.
		If ifft was not called first, inaccurate results might happen. It will be fixed later on.
		Check calculate function's docstring for more detail.
		"""
		if not self._ifft_called_first:
			warnings.warn('This module is designed to call ifft before fft, so inconsistencies might occur when calling fft first. Consider using numpys fft package with your own logic. This functionality will be added later on.', FourierWarning)
		self.x, self.y = fft_method(self.original_x, self.y)

	def window(self, at, fwhm, window_order=6, plot=True):
		"""
		Draws a gaussian window on the plot with the desired parameters.
		The maximum value is adjusted for the dataset mostly for visibility reasons.
		You should explicitly call self.show() after this function is set.

		Parameters:
		----------

		at: float
			maximum of the gaussian curve

		fwhm: float
			Full width at half maximum of the gaussian

		window_order: int, default is 6
			Order of the gaussian curve.
			If not even, it's incremented by 1.
		"""
		self.at = at
		self.fwhm = fwhm
		self.window_order = window_order
		gaussian = gaussian_window(self.x, self.at, self.fwhm, self.window_order)
		self.plotwidget.plot(self.x, gaussian*max(abs(self.y)), 'r--')
		if plot:
			self.show()

	def apply_window(self):
		"""
		If window function is correctly set, applies changes to the dataset.
		"""
		self.plotwidget.clf()
		self.plotwidget.cla()
		self.plotwidget.close()
		self.y = cut_gaussian(self.x, self.y, spike=self.at, fwhm=self.fwhm, win_order=self.window_order)
		
	@print_disp
	def calculate(self, reference_point, order, show_graph=False):
		""" 
		FFTMethod's calculate function.

		Parameters:
		----------

		reference_point: float
			reference point on x axis

		fit_order: int
			Polynomial (and maximum dispersion) order to fit. Must be in [1,5].

		show_graph: bool, optional
			shows a the final graph of the spectral phase and fitted curve.

		Returns:
		-------

		dispersion: array-like
			[GD, GDD, TOD, FOD, QOD]

		dispersion_std: array-like
			standard deviations due to uncertanity of the fit
			[GD_std, GDD_std, TOD_std, FOD_std, QOD_std]

		fit_report: lmfit report
			if lmfit is available, the fit report

		Notes:
		------

		Decorated with print_disp, so the results are immediately printed without explicitly saying so.

		Currently the x-axis transformation is sloppy, because we cache the 
		original x axis and not transforming it	backwards. In addition we need to keep track 
		of interpolation and zero-padding too. Currently the transforms are correct only if 
		first ifft was used. For now it's doing okay: giving good results. 
		For consistency we should still implement that a better way later.
		"""
		dispersion, dispersion_std, fit_report = args_comp(
			self.x, self.y, ref_point=reference_point, fit_order=order, show_graph=show_graph
			)
		self._dispersion_array = dispersion
		return dispersion, dispersion_std, fit_report


	def autorun(
		self, reference_point=None, order=None, *, enable_printing=True,
		skip_domain_check=False, only_phase=False, show_graph=True,
		usenifft=False
		):

		if not reference_point or not order:
			only_phase = True

		if not enable_printing:

			with open(os.devnull, "w") as g, contextlib.redirect_stdout(g):
				_run(self, skip_domain_check=skip_domain_check, show_graph=show_graph, usenifft=usenifft)
			if only_phase:
				y = np.unwrap(np.angle(self.y), axis=0)
				self.phase = Phase(self.x, y)
				self.phase.plot()
				return self.x, y
			self.calculate(reference_point=reference_point, order=order, show_graph=True)
		else:
			_run(self, skip_domain_check=skip_domain_check, show_graph=show_graph, usenifft=usenifft)
			if only_phase:
				y = np.unwrap(np.angle(self.y), axis=0)
				self.phase = Phase(self.x, y)
				self.phase.plot()
				return self.x, y
			self.calculate(reference_point=reference_point, order=order, show_graph=True)


import warnings

warnings.filterwarnings("ignore", message="invalid value encountered in sqrt")
warnings.filterwarnings("ignore", message="divide by zero encountered in true_divide")


from pysprint.utils import run_from_ipython
# setting up the IPython notebook
if run_from_ipython():
	from matplotlib import pyplot as plt
	plt.rcParams['figure.figsize'] = [15, 5]

__version__ = '0.7.0'
__author__ = 'Leéh Péter'

from .api import *







from functools import wraps
from math import factorial

import numpy as np
from scipy.interpolate import interp1d
import scipy.stats as st

__all__ = ['unpack_lmfit', 'find_nearest',
           '_handle_input', 'print_disp', 'fourier_interpolate',
           'between', 'get_closest', 'run_from_ipython',
           'calc_envelope', 'measurement', '_maybe_increase_before_cwt',
           'pad_with_trailing_zeros']


def _maybe_increase_before_cwt(y, tolerance=0.05):
    y = np.asarray(y)
    value = y.min()
    if value - tolerance >= 0:
        return False
    if value < 0 and np.abs(value) - tolerance > 0:
        return True
    return True


def calc_envelope(x, ind, mode='u'):
    '''Source: https://stackoverflow.com/a/39662343/11751294
    '''
    x_abs = np.abs(x)
    if mode == 'u':
        loc = np.where(np.diff(np.sign(np.diff(x_abs))) < 0)[0] + 1
    elif mode == 'l':
        loc = np.where(np.diff(np.sign(np.diff(x_abs))) > 0)[0] + 1
    else:
        raise ValueError('mode must be u or l.')
    peak = x_abs[loc]
    envelope = np.interp(ind, loc, peak)
    return envelope, peak, loc

def run_from_ipython():
    try:
        __IPYTHON__
        return True
    except NameError:
        return False

def get_closest(xValue, xArray, yArray):
    idx = (np.abs(xArray - xValue)).argmin()
    value = xArray[idx]
    return value, yArray[idx], idx

def between(val, except_around):
    if except_around is None:
        return False
    elif len(except_around) != 2:
        raise ValueError(f'Invalid interval. Try [start, end] instead of {except_around}')
    else:
        lower = float(min(except_around))
        upper = float(max(except_around))
    if val <= upper and val >= lower:
        return True
    return False


def unpack_lmfit(r):
    dispersion, dispersion_std = [], []
    for name, par in r:
        dispersion.append(par.value)
        dispersion_std.append(par.stderr)
    return dispersion, dispersion_std


def measurement(array, confidence=0.95, silent=False):
    """
    Give the measurement results with condifence interval
    assuming the standard deviation is unknown.

    Parameters
    ----------

    array : ndarray
        The array containing the measured values

    confidence : float, optional
        The desired confidence level. Must be between 0 and 1.

    Returns
    -------

    mean: float
        The mean of the given array

    conf: tuple-like (interval)
        The confidence interval

    Example(s)
    ---------
    >>> import numpy as np
    >>> from pysprint.utils import measurement
    >>> a = np.array([123.783, 121.846, 122.248, 125.139, 122.569])
    >>> mean, interval = measurement(a, 0.99)
    123.117000 ± 2.763022
    >>> mean
    123.117
    >>> interval
    (120.35397798230359, 125.88002201769642)

    Notes
    -----

    I decided to print the results immediately, because people often don't use
    it for further code. Of course, they are also returned if needed.

    """
    mean = np.mean(array)
    conf = st.t.interval(confidence, len(array)-1, loc=mean, scale=st.sem(array))
    if not silent:
        print(f'{mean:5f} ± {(mean-conf[0]):5f}')
    return mean, conf


def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(value - array)).argmin()
    return array[idx], idx


def _handle_input(x, y, ref, sam):
    """
    Instead of handling the inputs in every function, there is this private method.

    Parameters
    ----------

    x: array-like
    x-axis data

    y: array-like
    y-axis data

    ref, sam: array-like
    reference and sample arm spectrum evaluated at x

    Returns
    -------
    x: array-like
    unchanged x data

    Ydata: array-like
    the transformed y data

    """
    if len(x) == 0:
        raise ValueError('No values for x.')

    if len(y) == 0:
        raise ValueError('No values for y.')

    if (len(x) > 0) and (len(ref) > 0) and (len(sam) > 0):
        y_data = (y - ref - sam) / (2 * np.sqrt(ref * sam))
    elif (len(ref) == 0) or (len(sam) == 0):
        y_data = y
    else:
        raise TypeError('Input types are wrong.\n')
    return np.asarray(x),  np.asarray(y_data)


def print_disp(f):
    @wraps(f)
    def wrapping(*args, **kwargs):
        disp, disp_std, st = f(*args, **kwargs)
        labels = ('GD', 'GDD','TOD', 'FOD', 'QOD', 'SOD')
        disp = np.trim_zeros(disp, 'b')
        disp_std = disp_std[:len(disp)]
        for i, (label, disp_item, disp_std_item) in enumerate(zip(labels, disp, disp_std)):
            print(f'{label} = {disp_item:.5f} ± {disp_std_item:.5f} fs^{i+1}')
        return disp, disp_std, st
    return wrapping


def fourier_interpolate(x, y):
    ''' Simple linear interpolation for FFTs'''
    xs = np.linspace(x[0], x[-1], len(x))
    intp = interp1d(x, y, kind='linear', fill_value='extrapolate')
    ys = intp(xs)
    return xs, ys

def pad_with_trailing_zeros(array, shape):
    '''
    Pad an array with trailing zeros to be the desired shape
    '''
    c = shape - len(array)
    return np.pad(array, pad_width=(0, c), mode='constant')



"""
Methods for manipulating the loaded data
"""
import numpy as np
from scipy.signal import find_peaks, savgol_filter, gaussian, convolve, find_peaks_cwt
from scipy.interpolate import interp1d
from pysprint.utils import find_nearest, _handle_input, between, _maybe_increase_before_cwt


def cwt(x, y, ref, sam, width, floor_thres=0.1):
	x, y = _handle_input(
		x, y, ref, sam
		)
	idx = find_peaks_cwt(y, np.arange(1, width))
	if _maybe_increase_before_cwt(y, tolerance=floor_thres):
		y += 2
	y_rec = 1/y
	idx2 = find_peaks_cwt(y_rec, np.arange(1, width))
	return x[idx], y[idx]-2, x[idx2], y[idx2]-2

def savgol(x, y, ref, sam, window=101, order=3):
	x, y = _handle_input(
		x, y, ref, sam
		)
	xint, yint = interpolate_data(x, y, [], [])
	if window > order:
		try:
			if window % 2 == 1:
				fil = savgol_filter(yint, window_length=window, polyorder=order)
				return xint, fil
			else:
				fil = savgol_filter(yint, window_length=window + 1, polyorder=order)
				return xint, fil
		except Exception as e:
			print(e)
	else:
		raise ValueError(f'Order must be lower than window length. Currently window is {window} and order is {order})')

def find_peak(x, y, ref, sam, proMax=1, proMin=1, threshold=0.1, except_around=None):   
	if except_around is not None and len(except_around) != 2:
		raise ValueError('Invalid except_around arg. Try [start, stop].')
	if except_around is not None:
		try:
			float(except_around[0])
			float(except_around[1])
		except ValueError:
			raise ValueError(f'Invalid except_around arg. Only numeric values are allowed.')
	x, y = _handle_input(x, y, ref, sam)
	maxIndexes, _ = find_peaks(y, prominence=proMax) 
	y_rec = 1/y
	minIndexes, _ = find_peaks(y_rec, prominence=proMin)
	min_idx = []
	max_idx = []
	for idx in maxIndexes:
		if between(x[idx], except_around) or np.abs(y[idx]) > threshold:
			max_idx.append(idx)
	for idx in minIndexes:
		if between(x[idx], except_around) or np.abs(y[idx]) > threshold:
			min_idx.append(idx)

	if len(x[max_idx]) != len(y[max_idx]) or len(x[min_idx]) != len(y[min_idx]):
		raise ValueError('Something went wrong, try to cut the edges of data.')

	return x[max_idx], y[max_idx], x[min_idx], y[min_idx]

def convolution(x, y, ref, sam,	win_len, standev=200):
	x, y = _handle_input(
		x, y, ref, sam
		)
	if win_len < 0 or win_len > len(x):
		raise ValueError('Window length must be 0 < window_length < len(x)')

	xint, yint = interpolate_data(x, y, [], [])
	window = gaussian(win_len, std=standev)
	smoothed = convolve(yint, window/window.sum(), mode='same')
	return xint, smoothed

def interpolate_data(x, y, ref, sam):
	x, y = _handle_input(x, y, ref, sam)
	xint = np.linspace(x[0], x[-1], len(x))
	intp = interp1d(x,y, kind='linear')
	yint = intp(xint)
	return xint, yint


def cut_data(x, y, ref, sam, start=None, stop=None):
	x, y = _handle_input(x, y, ref, sam)
	if start is None:
		start = np.min(x)
	if stop is None:
		stop = np.max(x)
	if start < stop:
		low_item, _ = find_nearest(x, start)
		high_item, _ = find_nearest(x, stop)
		mask = np.where((x >= low_item) & (x <= high_item))
		return x[mask], y[mask]
	elif stop < start:
		raise ValueError('Start must not exceed stop value.')
	else:
		return np.array([]), np.array([])



import os
import contextlib
import warnings

import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import find_peaks


def signaltonoise(a, axis=0, ddof=0):
    a = np.asanyarray(a)
    m = a.mean(axis)
    sd = a.std(axis=axis, ddof=ddof)
    return np.where(sd == 0, 0, m/sd)


def find_roi(x, y):
    y = np.abs(y)
    idx = np.where(x > 0)
    return x[idx], y[idx]


def find_center(x, y, n_largest=5):

    peaks, props = find_peaks(y, prominence=0.001, height=np.max(y) / 100)

    if n_largest > len(props['prominences']):
        n_largest = len(props['prominences'])

    # find the most outlying peaks from the noise
    ind1 = np.argpartition(props['prominences'], -n_largest)[-n_largest:]
    # find the highest peaks by value
    ind2 = np.argpartition(props['peak_heights'], -n_largest)[-n_largest:]

    ind = np.unique(np.concatenate((ind1, ind2)))
    
    peaks = peaks[ind]
    _x, _y = x[peaks], y[peaks]
    residx = np.argmax(_x * _y) # weighted with distance from origin

    return _x[residx], _y[residx]

def _ensure_window_at_origin(center, fwhm, order, peak_center_height, tol=1E-3):
    '''
    Ensure that the gaussian window of given parameters is
    not crossing zero.
    '''
    std = fwhm/(2 * (np.log(2)**(1 / order)))
    val = np.exp(-((0-center)**order)/(std**order))
    return val < peak_center_height * tol, val


def predict_fwhm(x, y, center, peak_center_height, prefer_high_order=True, tol=1E-3):
    if np.iscomplexobj(y):
        y = np.abs(y)

    N = len(x)

    peak_position = center/N


    # explicitly prefer higher order windows if the peak is too close to the origin

    if peak_position < 0.2:
        warnings.warn('The peak is too close to the origin, manual control is advised.', UserWarning)
        prefer_high_order = True

    if prefer_high_order:
        order_choices = (8, 10, 12)
    else:
        order_choices = (2, 4, 6)

    sgnl = signaltonoise(y)
    if peak_position < 0.6:
        if sgnl > 0.1:
            center += N * 0.05 + (sgnl-0.1) * N * 2
            window_size = 0.75 * center * 2 + (sgnl-0.1) * N * 0.2
        else:
            center += N * 0.05
            window_size = 0.75 * center * 2
    else:
        if sgnl > 0.1:
            center += N * 0.05 + (sgnl-0.1) * N * 2
            window_size = 0.5 * center * 2 + (sgnl-0.1) * N * 0.2
        else:
            center += N * 0.05
            window_size = 0.5 * center * 2
    try:
        order = order_choices[int(2-(3*peak_position//2))]
    except IndexError:
        order = order_choices[0]

    if _ensure_window_at_origin(center, window_size, order, peak_center_height, tol=tol)[0]:
        return center, window_size, order
    else:
        val = _ensure_window_at_origin(center, window_size, order, peak_center_height, tol=tol)[1]
        # print(f'The window is bigger at the origin than the desired tolerance. Actual: {val:.4e}, Desired: {tol*peak_center_height:.4e}')
        warnings.warn(f'The window is bigger at the origin than the desired tolerance. Actual:{val:.4e}, Desired:{tol*peak_center_height:.4e}')
        return center, window_size, order + 2

def _run(ifg, skip_domain_check=False, only_phase=False, show_graph=True, usenifft=False):
    print('Interferogram received.')
    if ifg.probably_wavelength is True and not skip_domain_check:
        print('Probably in wavelength domain, changing to frequency...', end='', flush=True)
        ifg.chdomain()
        print('Done')
    if usenifft:
        pprint = ' using NUIFFT algorithm...'
    else:
        pprint = '...'
    print(f'Applying IFFT{pprint}', end='', flush=True)
    ifg.ifft(usenifft=usenifft)
    print('Done')
    x, y =  find_roi(ifg.x, ifg.y)
    a, b = find_center(x, y)
    print('Acquiring gaussian window parameters...', end='', flush=True)
    print('Done')
    c, ws, _order = predict_fwhm(x, y, center=a, peak_center_height=b)
    print(f'A {_order} order gaussian window centered at {c:.2f} fs with FWHM {ws:.2f} fs was applied.')
    ifg.window(at=c, fwhm=ws, window_order=_order, plot=show_graph)
    ifg.apply_window()
    print(f'Applying FFT...', end='', flush=True)
    ifg.fft()
    print('Done')
    print('Calculating...')
